# -*- coding: utf-8 -*-
from bda.plone.orders.common import get_order
from bda.plone.payment import Payment
from bda.plone.payment import Payments
from Products.Five import BrowserView
from zope.i18nmessageid import MessageFactory


_ = MessageFactory("bda.plone.payment")


class DebitOrder(Payment):
    pid = "debit_order"
    label = _("debit_order", default=u"Debit Order")

    def init_url(self, uid):
        return "%s/@@debit_order?uid=%s" % (self.context.absolute_url(), uid)


class DoDebitOrder(BrowserView):
    def __call__(self, **kw):
        uid = self.request["uid"]
        payment = Payments(self.context).get("debit_order")
        payment.succeed(self.request, uid)
        url = "%s/@@debit_ordered?uid=%s" % (self.context.absolute_url(), uid)
        self.request.response.redirect(url)


class DebitOrderFinished(BrowserView):
    def id(self):
        uid = self.request.get("uid", None)
        try:
            order = get_order(self.context, uid)
        except ValueError:
            return None
        return order.attrs.get("ordernumber")



# -*- coding: utf-8 -*-
# TODO: fix dependency on bda.plone.orders.
# this invalidates the dependency chain.
from bda.plone.orders.common import get_order
from bda.plone.payment import Payment
from bda.plone.payment import Payments
from plone.protect.authenticator import createToken
from Products.Five import BrowserView
from zope.i18nmessageid import MessageFactory
from zope.interface import Attribute
from zope.interface import Interface


_ = MessageFactory("bda.plone.payment")


class ICashOnDeliverySettings(Interface):

    currency = Attribute(u"Currency of cash on delivery")

    costs = Attribute(u"Costs of cash on delivery as decimal in gross")


class CashOnDelivery(Payment):
    pid = "cash_on_delivery"

    @property
    def label(self):
        settings = ICashOnDeliverySettings(self.context)
        return _(
            "cash_on_delivery",
            default=u"Cash on delivery. An extra fee of ${costs} "
            u"${currency} will be charged on order delivery",
            mapping={"costs": settings.costs, "currency": settings.currency},
        )

    def init_url(self, uid):
        return "{url}/@@cash_on_delivery?uid={uid}&_authenticator={token}".format(
            url=self.context.absolute_url(), uid=uid, token=createToken()
        )


class DoCashOnDelivery(BrowserView):
    def __call__(self, **kw):
        uid = self.request["uid"]
        payment = Payments(self.context).get("cash")
        payment.succeed(self.request, uid)
        url = "{url}/@@cash_on_delivery_done?uid={uid}&_authenticator={token}".format(
            url=self.context.absolute_url(), uid=uid, token=createToken()
        )
        self.request.response.redirect(url)


class CashOnDeliveryFinished(BrowserView):
    def id(self):
        uid = self.request.get("uid", None)
        try:
            order = get_order(self.context, uid)
        except ValueError:
            return None
        return order.attrs.get("ordernumber")



# -*- coding: utf-8 -*-
from .. import Payment
from .. import Payments
from ..interfaces import IPaymentData
from plone import api
from plone.keyring.interfaces import IKeyManager
from plone.protect.authenticator import createToken
from plone.protect.utils import getRoot
from plone.protect.utils import getRootKeyManager
from Products.Five import BrowserView
from zExceptions import Redirect
from zope.component import ComponentLookupError
from zope.component import getUtility
from zope.i18nmessageid import MessageFactory
import base64
import json
import logging
import requests
import transaction
import uuid


logger = logging.getLogger("bda.plone.payment")
_ = MessageFactory("bda.plone.payment")


###############################################################################
# Account settings
USERNAME = ''
PASSWORD = ''
CUSTOMER_ID = ''
TERMINAL_ID = ''
CONFIG_SET = ''
TESTING = False
# End account settings
###############################################################################


class SaferPayError(Exception):
    """Raised if SIX payment return an error.
    """


class SaferPay(object):
    spec_version = '1.12'

    @property
    def testing(self):
        return TESTING

    @property
    def username(self):
        return USERNAME

    @property
    def password(self):
        return PASSWORD

    @property
    def customer_id(self):
        return CUSTOMER_ID

    @property
    def terminal_id(self):
        return TERMINAL_ID

    @property
    def config_set(self):
        return CONFIG_SET

    @property
    def base_url(self):
        if self.testing:
            return 'https://test.saferpay.com/api/Payment/v1'
        return 'https://www.saferpay.com/api/Payment/v1'

    @property
    def initialize_url(self):
        return '{}/PaymentPage/Initialize'.format(self.base_url)

    @property
    def assert_url(self):
        return '{}/PaymentPage/Assert'.format(self.base_url)

    @property
    def capture_url(self):
        return '{}/Transaction/Capture'.format(self.base_url)

    @property
    def cancel_url(self):
        return '{}/Transaction/Cancel'.format(self.base_url)

    @property
    def credentials(self):
        username = self.username
        password = self.password
        if not isinstance(username, bytes):
            username = username.encode('utf-8')
        if not isinstance(password, bytes):
            password = password.encode('utf-8')
        credentials = username + b':' + password
        return base64.b64encode(credentials)

    def request(self, url, data):
        """Response status codes:
            200 - OK (No error)
            400 - Validation error
            401 - Authentication of the request failed
            402 - Requested action failed
            403 - Access denied
            406 - Not acceptable (wrong accept header)
            415 - Unsupported media type (wrong content-type header)
            500 - Internal error

        If response content type 'application/json', an error message is
        present.
        """
        headers = {
            'Content-Type': 'application/json; charset=utf-8',
            'Accept': 'application/json',
            'Authorization': 'Basic {}'.format(self.credentials)
        }
        request_id = str(uuid.uuid4())
        data['RequestHeader'] = {
            'SpecVersion': self.spec_version,
            'CustomerId': self.customer_id,
            'RequestId': request_id,
            'RetryIndicator': 0
        }
        response = requests.post(url, headers=headers, data=json.dumps(data))
        if response.status_code != 200:
            if response.headers['Content-Type'].find('application/json') == -1:
                msg = 'An error occurred: {}'.format(response.status_code)
                raise SaferPayError(msg)
            rdata = response.json()
            msg = u'Error in response: {} {}'.format(
                rdata['ErrorName'],
                rdata['ErrorMessage']
            )
            raise SaferPayError(msg)
        rdata = response.json()
        header = rdata['ResponseHeader']
        if header['RequestId'] != request_id:
            msg = u'Unknown response: {} != {}'.format(
                header['RequestId'],
                request_id
            )
            raise SaferPayError(msg)
        return rdata

    def initialize(
        self,
        amount,
        currency,
        order_id,
        payment_description,
        success_link,
        fail_link
    ):
        data = {
            'TerminalId': self.terminal_id,
            'Payment': {
                'Amount': {
                    'Value': amount,
                    'CurrencyCode': currency
                },
                'OrderId': order_id,
                'Description': payment_description
            },
            'ReturnUrls': {
                'Success': success_link,
                'Fail': fail_link
            }
        }
        config_set = self.config_set
        if config_set:
            data['ConfigSet'] = config_set
        return self.request(self.initialize_url, data)

    def assert_(self, token):
        data = {
            'Token': token
        }
        return self.request(self.assert_url, data)

    def capture(self, tid):
        data = {
            'TransactionReference': {
                'TransactionId': tid
            }
        }
        return self.request(self.capture_url, data)

    def cancel(self, tid):
        data = {
            'TransactionReference': {
                'TransactionId': tid
            }
        }
        return self.request(self.cancel_url, data)


class URLMixin(object):

    def payment_link(self, base_url, view_name, order_uid):
        try:
            key_manager = getUtility(IKeyManager)
        except ComponentLookupError:
            key_manager = getRootKeyManager(getRoot(self.context))
        authenticator = createToken(manager=key_manager)
        return '{}/@@{}?uid={}&_authenticator={}'.format(
            base_url,
            view_name,
            order_uid,
            authenticator
        )


class SixPayment(Payment, URLMixin):
    pid = "six_payment"
    label = _("six_payment", "Six Payment")

    def init_url(self, uid):
        return self.payment_link(
            self.context.absolute_url(),
            'saferpay_initialize',
            uid
        )


class SaferPayBaseView(BrowserView, URLMixin):

    @property
    def shopmaster_mail(self):
        # This is a soft dependency indirection on bda.plone.shop
        name = "bda.plone.shop.interfaces.IShopSettings.admin_email"
        email = api.portal.get_registry_record(name=name, default=None)
        if email is not None:
            return email
        logger.warning("No shop master email was set.")
        return "(no shopmaster email set)"


class SaferPayInitialize(SaferPayBaseView):

    def __call__(self):
        base_url = self.context.absolute_url()
        order_uid = self.request['uid']
        success_link = self.payment_link(base_url, 'saferpay_assert', order_uid)
        fail_link = self.payment_link(base_url, 'saferpay_failed', order_uid)
        try:
            p_data = IPaymentData(self.context)
            data = p_data.data(order_uid)
            ordernumber = data['ordernumber']
            sp = SaferPay()
            response = sp.initialize(
                data["amount"],
                data["currency"],
                data["ordernumber"],
                data["description"],
                success_link,
                fail_link
            )
            annotations = p_data.annotations(ordernumber)
            annotations['saferpay_token'] = response['Token']
            redirect_url = response['RedirectUrl']
            transaction.commit()
        except Exception as e:
            logger.error("Cannot initialize payment: '{}'".format(e))
            redirect_url = fail_link
        raise Redirect(redirect_url)


class SaferPayAssert(SaferPayBaseView):

    def __call__(self):
        base_url = self.context.absolute_url()
        order_uid = self.request['uid']
        success_link = self.payment_link(base_url, 'saferpay_success', order_uid)
        fail_link = self.payment_link(base_url, 'saferpay_failed', order_uid)
        try:
            p_data = IPaymentData(self.context)
            data = p_data.data(order_uid)
            ordernumber = data['ordernumber']
            annotations = p_data.annotations(ordernumber)
            token = annotations['saferpay_token']
            sp = SaferPay()
            response = sp.assert_(token)
            ta = response['Transaction']
            status = ta['Status']
            tid = ta['Id']
            annotations['saferpay_transaction_id'] = tid
            ordernumber = ta['OrderId']
            if status == 'PENDING':
                # Currently only paydirect
                msg = 'Paydirect not supported yet'
                raise NotImplementedError(msg)
            elif status == 'AUTHORIZED':
                captured_response = sp.capture(tid)
                c_status = captured_response['Status']
                if c_status == 'PENDING':
                    # Currently only paydirect
                    msg = 'Paydirect not supported yet'
                    raise NotImplementedError(msg)
                elif c_status == 'CAPTURED':
                    capture_id = captured_response['CaptureId']
                    annotations['saferpay_capture_id'] = capture_id
                else:
                    msg = 'Unknown capture status: {}'.format(c_status)
                    raise SaferPayError(msg)
            elif status == 'CAPTURED':
                capture_id = ta['CaptureId']
                annotations['saferpay_capture_id'] = capture_id
            else:
                msg = 'Unknown transaction status: {}'.format(status)
                raise SaferPayError(msg)
            redirect_url = success_link
            transaction.commit()
        except Exception as e:
            logger.error("Cannot finalize payment: '{}'".format(e))
            redirect_url = fail_link
        raise Redirect(redirect_url)


class SaferPaySuccess(SaferPayBaseView):

    def finalize(self):
        order_uid = self.request['uid']
        p_data = IPaymentData(self.context)
        data = p_data.data(order_uid)
        ordernumber = data['ordernumber']
        annotations = p_data.annotations(ordernumber)
        tid = annotations['saferpay_transaction_id']
        evt_data = {'tid': tid}
        payment = Payments(self.context).get('six_payment')
        payment.succeed(self.request, order_uid, evt_data)


class SaferPayFailed(SaferPayBaseView):

    def finalize(self):
        payment = Payments(self.context).get('six_payment')
        payment.failed(self.request, self.request['uid'], {'tid': 'none'})



# -*- coding: utf-8 -*-
from bda.plone.orders.common import get_order
from bda.plone.payment import Payment
from bda.plone.payment import Payments
from plone.protect.authenticator import createToken
from Products.Five import BrowserView
from zope.i18nmessageid import MessageFactory


_ = MessageFactory("bda.plone.payment")


class Invoice(Payment):
    pid = "invoice"
    label = _("invoice", default=u"Invoice")

    def init_url(self, uid):
        return "{url}/@@pay_invoice?uid={uid}&_authenticator={token}".format(
            url=self.context.absolute_url(), uid=uid, token=createToken()
        )


class DoInvoice(BrowserView):
    def __call__(self, **kw):
        uid = self.request["uid"]
        payment = Payments(self.context).get("invoice")
        payment.succeed(self.request, uid)
        url = "{url}/@@pay_invoice_done?uid={uid}&_authenticator={token}".format(
            url=self.context.absolute_url(), uid=uid, token=createToken()
        )
        self.request.response.redirect(url)


class InvoiceFinished(BrowserView):
    def id(self):
        uid = self.request.get("uid", None)
        try:
            order = get_order(self.context, uid)
        except ValueError:
            return None
        return order.attrs.get("ordernumber")



# -*- coding: utf-8 -*-
from zope.interface import Attribute
from zope.interface import Interface


class IPaymentExtensionLayer(Interface):
    """Browser layer for bda.plone.payment.
    """


class IPaymentEvent(Interface):
    """Payment related event.
    """

    context = Attribute(u"Context in which this event was triggered.")

    request = Attribute(u"Current request.")

    payment = Attribute(u"Payment instance.")

    order_uid = Attribute(u"Referring order uid.")

    data = Attribute(u"Optional data as dict")


class IPaymentSuccessEvent(IPaymentEvent):
    """This event gets triggered when payment was successful.
    """


class IPaymentFailedEvent(IPaymentEvent):
    """This event gets triggered when payment failed.
    """


class IPaymentSettings(Interface):
    """Payment availability and default settings.
    """

    available = Attribute(u"List of available payment method ids")

    default = Attribute(u"Default payment method")


class IPayment(Interface):
    """Single payment.
    """

    pid = Attribute(
        u"Unique payment id. Payment adapter is also registered " u"under this name."
    )

    label = Attribute(u"Payment label")

    clear_session = Attribute(
        u"Clear cart immediately on checkout success. Defaults to True. "
        u"May be set to False in redirect payments. If so, cart session "
        u"clearance must be handled in payment success and failed event handlers."
    )

    available = Attribute(
        u"Flag whether payment is available in recent " u"payment cycle"
    )

    default = Attribute(u"Flag whether this payment is default payment.")

    def init_url(uid):
        """Return payment initialization URL.
        """

    def succeed(request, order_uid, data=dict()):
        """Notify ``IPaymentSuccessEvent``.
        """

    def failed(request, order_uid, data=dict()):
        """Notify ``IPaymentFailedEvent``.
        """


class IPaymentData(Interface):
    """Interface for providing data required by payment.
    """

    def uid_for(ordernumber):
        """Return order_uid for ordernumber.
        """

    def annotations(ordernumber):
        """Dict like object for annotating payment related data.
        """

    def data(order_uid):
        """Return dict containing payment related order data like:

        {
            'amount': '1000',
            'currency': 'EUR',
            'description': 'description',
            'ordernumber': '1234567890',
        }
        """



# -*- coding: utf-8 -*-
from bda.plone.payment.interfaces import IPayment
from bda.plone.payment.interfaces import IPaymentEvent
from bda.plone.payment.interfaces import IPaymentFailedEvent
from bda.plone.payment.interfaces import IPaymentSettings
from bda.plone.payment.interfaces import IPaymentSuccessEvent
from zope.component import adapter
from zope.component import getAdapter
from zope.component import getAdapters
from zope.event import notify
from zope.interface import implementer
from zope.interface import Interface


@implementer(IPaymentEvent)
class PaymentEvent(object):
    def __init__(self, context, request, payment, order_uid, data):
        self.context = context
        self.request = request
        self.payment = payment
        self.order_uid = order_uid
        self.data = data


@implementer(IPaymentSuccessEvent)
class PaymentSuccessEvent(PaymentEvent):
    pass


@implementer(IPaymentFailedEvent)
class PaymentFailedEvent(PaymentEvent):
    pass


class Payments(object):
    def __init__(self, context):
        self.context = context

    def get(self, name):
        return getAdapter(self.context, IPayment, name=name)

    @property
    def payments(self):
        adapters = getAdapters((self.context,), IPayment)
        return [_[1] for _ in adapters]

    @property
    def vocab(self):
        adapters = getAdapters((self.context,), IPayment)
        return [(_[0], _[1].label) for _ in adapters if _[1].available]

    @property
    def default(self):
        adapters = getAdapters((self.context,), IPayment)
        for name, payment in adapters:
            if payment.default:
                return name
        if adapters:
            return adapters[0][0]


@implementer(IPayment)
@adapter(Interface)
class Payment(object):
    pid = None
    label = None
    # deferred = False  # XXX: remove
    clear_session = True

    def __init__(self, context):
        self.context = context

    @property
    def available(self):
        settings = IPaymentSettings(self.context)
        return self.pid in settings.available

    @property
    def default(self):
        settings = IPaymentSettings(self.context)
        return self.pid == settings.default

    def succeed(self, request, order_uid, data=dict()):
        evt = PaymentSuccessEvent(self.context, request, self, order_uid, data)
        notify(evt)

    def failed(self, request, order_uid, data=dict()):
        evt = PaymentFailedEvent(self.context, request, self, order_uid, data)
        notify(evt)

    def init_url(self, uid):
        raise NotImplementedError(
            u"Abstract ``Payment`` does not implement " u"``init_url``"
        )



# -*- coding: utf-8 -*-
from bda.plone.payment.tests import Payment_INTEGRATION_TESTING
from bda.plone.payment.tests import set_browserlayer

import unittest


class TestPayment(unittest.TestCase):
    layer = Payment_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]
        set_browserlayer(self.request)

    def test_foo(self):
        self.assertEquals(1, 1)



# -*- coding: utf-8 -*-
from bda.plone.payment.interfaces import IPaymentExtensionLayer
from plone.app.testing import IntegrationTesting
from plone.app.testing import PLONE_FIXTURE
from plone.app.testing import PloneSandboxLayer
from zope.interface import alsoProvides


def set_browserlayer(request):
    """Set the BrowserLayer for the request.

    We have to set the browserlayer manually, since importing the profile alone
    doesn't do it in tests.
    """
    alsoProvides(request, IPaymentExtensionLayer)


class PaymentLayer(PloneSandboxLayer):
    defaultBases = (PLONE_FIXTURE,)

    def setUpZope(self, app, configurationContext):
        import bda.plone.payment

        self.loadZCML(package=bda.plone.payment, context=configurationContext)

    def setUpPloneSite(self, portal):
        self.applyProfile(portal, "bda.plone.payment:default")

    def tearDownZope(self, app):
        pass


Payment_FIXTURE = PaymentLayer()
Payment_INTEGRATION_TESTING = IntegrationTesting(
    bases=(Payment_FIXTURE,), name="Payment:Integration"
)



# -*- coding: utf-8 -*-
from setuptools import find_packages
from setuptools import setup

import os


version = "2.0b1"
shortdesc = "Payment"
longdesc = open(os.path.join(os.path.dirname(__file__), "README.rst")).read()
longdesc += open(os.path.join(os.path.dirname(__file__), "CHANGES.rst")).read()
longdesc += open(os.path.join(os.path.dirname(__file__), "LICENSE.rst")).read()


setup(
    name="bda.plone.payment",
    version=version,
    description=shortdesc,
    long_description=longdesc,
    classifiers=[
        "Framework :: Plone",
        "Framework :: Plone :: 6.0",
        "Framework :: Plone :: Addon",
        "Framework :: Zope",
        "Framework :: Zope :: 5",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    author="BlueDynamics Alliance",
    author_email="dev@bluedynamics.com",
    license="GNU General Public Licence",
    packages=find_packages("src"),
    package_dir={"": "src"},
    namespace_packages=["bda", "bda.plone"],
    include_package_data=True,
    zip_safe=False,
    install_requires=["setuptools", "Products.CMFPlone", "requests"],
    extras_require={"test": ["plone.app.testing"]},
)



# -*- coding:utf-8 -*-
from plone.base.interfaces import INonInstallable
from zope.interface import implementer


@implementer(INonInstallable)
class HiddenProfiles(object):
    def getNonInstallableProfiles(self):
        """Do not show on Plone's list of installable profiles.
        """
        return []

    def getNonInstallableProducts(self):
        """Do not show on QuickInstaller's list of installable products.
        """
        return []
